package net.refractions.udig.mapgraphic.internal;

import java.io.IOException;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.IServiceInfo;
import net.refractions.udig.catalog.ITransientResolve;
import net.refractions.udig.mapgraphic.MapGraphicFactory;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;

/**
 * Service implementation for map graphics
 */
public class MapGraphicService extends IService {

    /** Dummy url for a MapGraphic */
    public final static URL SERVICE_URL;
    static {
        URL tmp;
        try {
            tmp = new URL("file:///localhost/mapgraphic"); //$NON-NLS-1$
        } catch (MalformedURLException e) {
            tmp=null;
            e.printStackTrace();
        }
        SERVICE_URL=tmp;
    }

    /** MapGraphic resource children * */
    private volatile List<MapGraphicResource> members;

    /** info object * */
    private MapGraphicServiceInfo info;

    /**
     * Construct <code>MapGraphicService</code>.
     */
    public MapGraphicService() {
        // NoOp
    }

    /*
     * @see net.refractions.udig.catalog.IService#resolve(java.lang.Class,
     *      org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public <T> T resolve( Class<T> adaptee, IProgressMonitor monitor ) throws IOException {
        if (monitor == null)
            monitor = new NullProgressMonitor();

        if (adaptee == null) {
            throw new NullPointerException("No adaptor specified" ); //$NON-NLS-1$
        }        
        if (adaptee.isAssignableFrom(MapGraphicFactory.class)) {
            return adaptee.cast(MapGraphicFactory.getInstance());
        }
        if (adaptee.isAssignableFrom(ITransientResolve.class)) {
            return adaptee.cast(this);
        }
        return super.resolve(adaptee, monitor);
    }
    @Override
    public synchronized IServiceInfo getInfo( IProgressMonitor monitor ) throws IOException {
        if (info == null){
            info = new MapGraphicServiceInfo();
        }
        return info;
    }
    /*
     * @see net.refractions.udig.catalog.IService#members(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public List< ? extends IGeoResource> members( IProgressMonitor monitor ) throws IOException {

        if (members == null) {
            synchronized (MapGraphicFactory.getInstance()) {
                if (members == null) {
                    members = new ArrayList<MapGraphicResource>();
                    List<IConfigurationElement> graphics = MapGraphicFactory.getInstance()
                            .getMapGraphics();

                    if (graphics != null) {
                        for( IConfigurationElement graphic : graphics ) {
                            members.add(new MapGraphicResource(this, graphic));
                        }
                    }
                }
            }
        }

        return members;
    }

    /*
     * @see net.refractions.udig.catalog.IService#getConnectionParams()
     */
    @Override
    public Map<String, Serializable> getConnectionParams() {
        HashMap<String, Serializable> params = new HashMap<String, Serializable>();
        params.put(MapGraphicServiceExtension.KEY, SERVICE_URL);
        return params;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#canResolve(java.lang.Class)
     */
    public <T> boolean canResolve( Class<T> adaptee ) {
        return adaptee != null
                && (adaptee.isAssignableFrom(MapGraphicFactory.class)
                        || adaptee.isAssignableFrom(ITransientResolve.class) || super
                        .canResolve(adaptee));
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getStatus()
     */
    public Status getStatus() {
        return Status.CONNECTED;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getMessage()
     */
    public Throwable getMessage() {
        return null;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getIdentifier()
     */
    public URL getIdentifier() {
        return SERVICE_URL;
    }

    static class MapGraphicServiceInfo extends IServiceInfo {
        /*
         * @see net.refractions.udig.catalog.IServiceInfo#getTitle()
         */
        @Override
        public String getTitle() {
            return Messages.MapGraphicService_title; 
        }

        @Override
        public String getDescription() {
            return Messages.MapGraphicService_description; 
        }
    }
}