/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.mapgraphic.scalebar;

import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.text.MessageFormat;
import java.util.Locale;

import net.refractions.udig.mapgraphic.MapGraphic;
import net.refractions.udig.mapgraphic.MapGraphicContext;
import net.refractions.udig.mapgraphic.MapGraphicPlugin;
import net.refractions.udig.mapgraphic.internal.Messages;
import net.refractions.udig.ui.graphics.ViewportGraphics;

import org.geotools.geometry.DirectPosition2D;
import org.geotools.referencing.GeodeticCalculator;
import org.opengis.geometry.DirectPosition;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.TransformException;

import com.vividsolutions.jts.geom.Coordinate;

/**
 * Provides a decorator object that represents a typical Scalebar
 * 
 * @author Richard Gould
 * @since 0.6.0
 */
public class ScalebarMapGraphic implements MapGraphic {
	
	public static final int KILOMETRE = 1000;
	public static final int METRE = 1;
	
	static int[] niceIntegers = new int[] {
		0, 1, 2, 5, //zero is a trick - we don't want to display "0km", but this way we can easily downgrade to metres 
		10, 20, 50,
		100, 200, 500,
		1000, 2000, 5000,
		10000
	};
	
	static int closestInt(int value, final int[] range) {
		if (range == null) {
			throw new NullPointerException("Range cannot be null"); //$NON-NLS-1$
		}
		
		if (range.length == 0) {
			throw new IllegalArgumentException("Range cannot be a zero length array"); //$NON-NLS-1$
		}
		
		for (int i = 0; i < range.length; i++) {
			if (value < range[i]) {
				if (i == 0) {
					return range[0];
				}
				
				int dist1 = Math.abs(value-range[i-1]);
				int dist2 = Math.abs(range[i]-value);
				
				if (dist1 == dist2) {
					return range[i-1];
				}
				
				if (dist1 < dist2) {
					return range[i-1];
				} else {
					return range[i];
				}
			}
		}
		return range[range.length-1];
	}

    /*
     * @see net.refractions.udig.project.render.decorator.Decorator#draw(java.awt.Graphics2D, net.refractions.udig.project.render.RenderContext)
     */
    public void draw( MapGraphicContext context) {
    	int measurement = KILOMETRE;
    	ViewportGraphics graphics = context.getGraphics();
        
        Rectangle location = (Rectangle) context.getLayer().getStyleBlackboard().get(LocationStyleContent.ID);
        if (location == null) {
            location = LocationStyleContent.createDefaultStyle();
            context.getLayer().getStyleBlackboard().put(LocationStyleContent.ID, location);
        }
        
        /*
         * 1. find out the width of the map [display] in pixels
         */
        
        int displayWidth = context.getRenderManagerInternal().getMapDisplay().getWidth();
        int displayHeight = context.getRenderManagerInternal().getMapDisplay().getHeight();
        
        /*
         * 2. take a coordinate from the left side and the right side of 
         * the map, in the map projection (whatever that may be).
         */        
        Coordinate min = context.getViewportModelInternal().pixelToWorld(0, displayHeight/2);
        Coordinate max = context.getViewportModelInternal().pixelToWorld(displayWidth, displayHeight/2);
        
        /*
         * 3. back-project the coordinates from 2. to latlong (or, if 
         * they are already in latlong, leave them be)
         * 
         * GeodeticCalculator can handle this for us -rg
         */        
        
        /*
         * 4. calculate the distance between the two point, using the 
         * distance on a spheroid functions from the CRS packages
		 *
         * 5. now you have the width of the screen in pixels, and the 
         * width of the screen in meters, so you know the number of 
         * ground units per pixel        
         */                
        double distance;
        double unitsPerPixel = 0.0;
        int trueBarLength = 0;
        int nice = 0;
        
        if (min == null || max == null){
            distance = 0.0;
        }
        else {
            try {
            	CoordinateReferenceSystem mapCRS = context.getViewportModel().getCRS();
                GeodeticCalculator calc = new GeodeticCalculator(mapCRS);
                calc.setStartingGeographicPoint(new DirectPosition2D(min.x, min.y));
                calc.setDestinationPosition(new DirectPosition2D(max.x, max.y));
                distance = calc.getOrthodromicDistance();
                
                unitsPerPixel = distance/displayWidth;
//                System.out.println("Units per pixel: " + unitsPerPixel);
                
                int idealBarLength = location.width;
//                System.out.println("Ideal bar length: " + idealBarLength);
                
                double mIdealBarDistance = (unitsPerPixel*idealBarLength);
//                System.out.println("Ideal bar distance: " + mIdealBarDistance);
                nice = closestInt((int) mIdealBarDistance/1000, niceIntegers) * 1000;
                
                if (nice == 0) {
                	nice = closestInt((int) mIdealBarDistance, niceIntegers);
                	measurement = METRE;
                }                
                
//                System.out.println("Nice distance: " + nice);
                
                calc.setDestinationGeographicPoint(new DirectPosition2D(min.x, min.y));
//                System.out.println("Anchor: " + new DirectPosition2D(min.x, min.y));
                calc.setDirection(90, nice); //90 is east
                DirectPosition nicePoint = calc.getDestinationPosition();
//                System.out.println("Nice point: " + nicePoint);
                
                Point nicePixels = context.worldToPixel(
                		new Coordinate(nicePoint.getOrdinate(0), 
                					   nicePoint.getOrdinate(1)));
                
//                System.out.println("Nice pixels: " + nicePixels);
                
                trueBarLength = nicePixels.x;
                
//                System.out.println("True bar length: " + trueBarLength);
            }
            catch ( IllegalArgumentException outOfRange ){
                distance = 0.0;
			} catch (TransformException e) {
				MapGraphicPlugin.log("Transform Exception while transforming positions", e); //$NON-NLS-1$
				distance = 0.0;
			}
        }
        if (distance == 0.0 ){
            graphics.setColor(Color.GRAY);            
            graphics.fillRect( location.x, location.y, location.width, location.height );
            graphics.setColor(Color.BLACK);            
            //graphics.drawRect( location.x, location.y, location.width, location.height );
            graphics.draw( new Rectangle(location.x, location.y, location.width, location.height ));
            
            int yText = (location.y < (displayHeight/2)) ? location.y+location.height+15 : location.y-15;
            graphics.drawString(Messages.ScalebarMapGraphic_zoomInRequiredMessage, location.x, yText, -1, -1); 
            
            return;
        }
        
        /*
         * Draw the scale bar
         */        
        int x = location.x;
        int y = location.y;
        
        int width = trueBarLength;
        int height = location.height;
        
        int barWidth = 4;
        
        int barStartX = x + barWidth/2;
        int barStartY = y;
        
        GeneralPath path = new GeneralPath();
        path.moveTo(barStartX, barStartY);
        path.lineTo(barStartX, barStartY + height);
        path.lineTo(barStartX+width, barStartY + height);
        path.lineTo(barStartX+width, barStartY);
                
        graphics.setColor(Color.WHITE);
        graphics.setStroke(ViewportGraphics.LINE_SOLID, barWidth+2);

        graphics.draw(path);
        
        graphics.setColor(Color.BLACK);
        graphics.setStroke(ViewportGraphics.LINE_SOLID, barWidth);

        graphics.draw(path);
        
        
        
    	MessageFormat formatter = new MessageFormat("", Locale.getDefault()); //$NON-NLS-1$
    	if (measurement == KILOMETRE) {
    		formatter.applyPattern(Messages.ScalebarMapGraphic_numberOfKilometres);
    	} else {
    		formatter.applyPattern(Messages.ScalebarMapGraphic_numberOfMetres);
    	}

    	Object[] arguments = { nice/measurement };
        String msg = formatter.format(arguments);
        
        int yText = y + (height+barWidth)/2 - 2;
        int xText = x + barWidth+2;

        //Draw a little alpha box underneath the text, so it can be viewed in dark areas.
        graphics.setColor(new Color(255, 255, 255, 103)); // 103/255 = ~40% transparency
        
        Rectangle2D msgBounds = graphics.getStringBounds(msg);
        RoundRectangle2D roundBounds = new RoundRectangle2D.Double();
        roundBounds.setRoundRect(
        		xText-1, yText-(int)msgBounds.getHeight()+3, 
        		msgBounds.getWidth()+4, msgBounds.getHeight()-1, 
        		10, 10);
        graphics.fill(roundBounds);
        
        //Draw the string denoting kilometres (or metres)
        graphics.setColor(Color.BLACK);
        graphics.drawString(msg, xText, yText, ViewportGraphics.ALIGN_LEFT, ViewportGraphics.ALIGN_BOTTOM);      
    }

}
