/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.internal;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.ICatalog;
import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.IServiceFactory;
import net.refractions.udig.catalog.ServiceParameterPersister;
import net.refractions.udig.catalog.URLUtils;
import net.refractions.udig.project.internal.impl.LayerImpl;
import net.refractions.udig.ui.ProgressManager;

import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.IExportedPreferences;
import org.eclipse.core.runtime.preferences.IPreferencesService;
import org.osgi.service.prefs.BackingStoreException;
import org.osgi.service.prefs.Preferences;

/**
 * Persists the connection parameters of the services that the resource needs for its resources.  
 * This is done so that if a map is sent to a friend the services can be created from the data provided by
 * the map alone.
 * 
 * @author Jesse
 * @since 1.0.0
 */
public class CatalogRef {
    protected Layer layer;


    protected static final String ENCODING = ProjectPlugin.Implementation.ENCODING;

	private static final String ERROR_SAVING = "Error saving"; //$NON-NLS-1$

    protected Map<URL, Map<String, Serializable>> connectionParams = Collections.synchronizedMap(new HashMap<URL, Map<String, Serializable>>());

    private volatile boolean loaded = false;

    /**
     * Construct <code>LayerRef</code>.
     * 
     * @param layer
     */
    public CatalogRef( Layer layer ) {
        this.layer = layer;
    }

    /**
     * Construct <code>LayerRef</code>.
     */
    public CatalogRef() {
        // do nothing
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        try{
        	List<IGeoResource> resources = layer.getGeoResources();
        	Preferences toSave = Platform.getPreferencesService().getRootNode().node(ProjectPlugin.ID).node(layerIDToString()); 
	        if (resources != LayerImpl.NULL) {
	            connectionParams.clear();
	
	            ServiceParameterPersister persister=new LayerCatalogRefPersister(connectionParams, getMapFile());
	         
	            persister.store(ProgressManager.instance().get(), toSave, resources);
	        }
	        
	        ByteArrayOutputStream out=new ByteArrayOutputStream();
	        Platform.getPreferencesService().exportPreferences((IEclipsePreferences) toSave, out, null);
	        toSave.clear();
	        
	        return URLEncoder.encode(out.toString(), ENCODING);
        }catch( Throwable t ){
        	ProjectPlugin.log("Error saving parameters for layer: "+(layer==null?"unkown id":layer.getID())+" this map cannot be sent to a collegue",t);  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$
        	return ERROR_SAVING;
        }
    }

	private File getMapFile() {
		return new File(layer.getMapInternal().eResource().getURI().toFileString());
	}

	private String layerIDToString() throws UnsupportedEncodingException {
		if( layer!=null && layer.getID()!=null )
			return URLEncoder.encode(layer.getID().toString(), ENCODING);
		else
			return "ID_"+System.currentTimeMillis(); //$NON-NLS-1$
	}

    /**
     * Reaods the parameters for the resource's parameters from the string and stores them.  
     * Services aren't created until load() is called.
     * 
     * @param string
     */
    public void parseResourceParameters( String string ) {
    	if( string==null || string.length()==0 ){
    		return;
    	}
    	String decoded;
    	try {
			decoded = URLDecoder.decode(string, ENCODING);
		} catch (UnsupportedEncodingException e) {
			decoded=string;
		}
		
		ByteArrayInputStream input=new ByteArrayInputStream(decoded.getBytes());
		try {
			IPreferencesService preferencesService = Platform.getPreferencesService();
			IExportedPreferences paramsNode = preferencesService.readPreferences(input);
			
			ServiceParameterPersister persister=new LayerCatalogRefPersister(connectionParams, null); //$NON-NLS-1$
			
			persister.restore(findParameterNode(paramsNode));
		} catch (Throwable e) {
			// ok maybe it is an from an older version of uDig so try the oldCatalogRef
			try{
				OldCatalogRef old=new OldCatalogRef();
				old.parseResourceParameters(string);
				connectionParams=old.connectionParams;
			}catch(Throwable e2){
				ProjectPlugin.log("CatalogRef#parseResourceParameters, couldn't load paramters", e); //$NON-NLS-1$
			}
		}
    	
    }

    private Preferences findParameterNode(IExportedPreferences paramsNode) throws BackingStoreException {
    	String[] name = paramsNode.childrenNames();
    	
    	Preferences plugin = paramsNode.node(name[0]);
    	name = plugin.childrenNames();
    	
    	return plugin.node(name[0]);
	}

	/**
     * @return Returns the layer.
     * @uml.property name="layer"
     */
    public Layer getLayer() {
        return layer;
    }

    /**
     * @param layer The layer to set.
     * @uml.property name="layer"
     */
    public void setLayer( Layer layer ) {
        this.layer = layer;
    }

    /**
     * Adds the required services into the catalog.
     */
    public void load() {
        ICatalog catalog = CatalogPlugin.getDefault().getLocalCatalog();
        IServiceFactory serviceFactory = CatalogPlugin.getDefault().getServiceFactory();
        synchronized (connectionParams) {
            for( Map.Entry<URL, Map<String, Serializable>> entry : connectionParams.entrySet() ) {
                Map<String, Serializable> params = entry.getValue();
                resolveURLs(params);
                if (catalog.getById(IService.class, entry.getKey(), ProgressManager.instance()
                        .get()) == null) {
                    for( IService service : serviceFactory
                            .acquire(entry.getKey(), entry.getValue()) ) {
                        catalog.add(service);
                    }
                }
            }
        }
        loaded = true;
    }

    private void resolveURLs( Map<String, Serializable> params ) {
        Set<Entry<String, Serializable>> entries = params.entrySet();
        for( Entry<String, Serializable> entry : entries ) {
            if( entry.getValue() instanceof URL ){
                String url = ((URL)entry.getValue()).toExternalForm();
                try {
                    entry.setValue(URLUtils.constructURL(getMapFile(), url ));
                } catch (MalformedURLException e) {
                    // oh well we'll keep see what happens, maybe this isn't possible
                    ProjectPlugin.log("Couldn't resolve the URL", e); //$NON-NLS-1$
                }
            }
        }
        
    }

    /**
     * Returns true if the required GeoResources have been added to the catalog.
     * 
     * @return
     * @uml.property name="loaded"
     */
    public boolean isLoaded() {
        return loaded;
    }
    
    private static class LayerCatalogRefPersister extends ServiceParameterPersister{

		private Map<URL, Map<String, Serializable>> allParams;

		public LayerCatalogRefPersister(Map<URL, Map<String, Serializable>> allParams, File mapFile) {
			super(CatalogPlugin.getDefault().getLocalCatalog(), CatalogPlugin.getDefault().getServiceFactory(), mapFile );
			this.allParams=allParams;
		}
		
		@Override
		protected void locateService(URL url, Map<String, Serializable> map) {
			if( allParams.containsKey(url) )
				ProjectPlugin.log("LayerCatalogRefPersister#locateService: duplicate resource ids when loading paramers"); //$NON-NLS-1$
			
			allParams.put(url, map);
		}
    	
    }
}
