/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.internal.render.impl;

import java.awt.Graphics2D;
import java.text.MessageFormat;
import java.util.Collection;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.internal.Messages;
import net.refractions.udig.project.internal.ProjectPlugin;
import net.refractions.udig.project.internal.render.ExecutorVisitor;
import net.refractions.udig.project.internal.render.RenderExecutor;
import net.refractions.udig.project.internal.render.Renderer;
import net.refractions.udig.project.render.IRenderContext;
import net.refractions.udig.project.render.IRenderer;
import net.refractions.udig.project.render.RenderException;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.common.notify.Notification;
import org.geotools.geometry.jts.ReferencedEnvelope;

import com.vividsolutions.jts.geom.Envelope;

/**
 * An Executor specifically for executing CompositeRenderers
 * 
 * @author jeichar
 * @since 0.6.0
 */
public class RenderExecutorComposite extends RenderExecutorMultiLayer {

    int timeout = 0;

    /**
     * Listens to the render events
     * 
     * @author jeichar
     * @since 0.6.0
     */
    protected static class CompositeRendererListener extends MultiLayerRendererListener {

        /**
         * Construct <code>CompositeRendererListener</code>.
         * 
         * @param executor
         */
        CompositeRendererListener( RenderExecutorComposite executor ) {
            super(executor);
        }

        RenderExecutorComposite getExecutor() {
            return (RenderExecutorComposite) executor;
        }

        /**
         * @see net.refractions.udig.project.internal.render.impl.RenderExecutorMultiLayer.MultiLayerRendererListener#stateChanged(org.eclipse.emf.common.notify.Notification)
         */
        protected void stateChanged( Notification msg ) {
            switch( msg.getNewIntValue() ){
            case Renderer.RENDERING:
                getExecutor().resetTimeout();
                synchronized (getExecutor()) {
                    // I think I want to do an update.  
                    // lets make it so if a renderer says update then we do it.
                    getExecutor().setState(RENDERING);
                }
                break;
            case Renderer.DONE:
                try {
                    synchronized (getExecutor()) {
                        if( !getExecutor().inUpdate ){
                            getExecutor().getRenderer().refreshImage();
                            executor.setState(Renderer.DONE);
                        }
                        if( !isRendering(getExecutor().getRenderer()) || getExecutor().renderJob.getMonitor().isCanceled() ){
                            getExecutor().notifyAll();
                        }
                    }
                    // state will be set in next else statement
                } catch (RenderException e) {
                    // won't happen.
                    ProjectPlugin.log("", e); //$NON-NLS-1$
                }
                break;
            case Renderer.RENDER_REQUEST:
                boolean oldValue = executor.eDeliver();
                executor.eSetDeliver(false);
                executor.setState(msg.getNewIntValue());
                executor.eSetDeliver(oldValue);
                ((RenderExecutorComposite) executor).refresh();
                break;
            case Renderer.STARTING:
                boolean oldValue2 = executor.eDeliver();
                executor.eSetDeliver(false);
                executor.setState(msg.getNewIntValue());
                executor.eSetDeliver(oldValue2);
                break;
            }
        }

    }

    private boolean inUpdate=false;
    
    protected static class CompositeRendererJob extends RenderJob {

      private static final int TIMEOUT = 20000;

        /**
         * Construct <code>CompositeRendererJob</code>.
         * 
         * @param executor
         */
        public CompositeRendererJob( RenderExecutorComposite executor ) {
            super(executor);
            setPriority(Job.INTERACTIVE);
        }
        
        @Override
        protected void init() {
        }

        /**
         * @see net.refractions.udig.project.internal.render.impl.RenderJob#getExecutor()
         */
        public RenderExecutorComposite getExecutor() {
            return (RenderExecutorComposite) executor;
        }

        /**
         * @see net.refractions.udig.project.internal.render.impl.RenderJob#startRendering()
         */
        protected void startRendering( Envelope bounds, IProgressMonitor monitor ) throws Throwable {
            // need to show that we are in update just in case a renderer triggers a state event
            // in the RenderExecutor.render() method.
            synchronized (getExecutor()) {
                getExecutor().inUpdate=true;
            }
            super.startRendering(bounds, monitor);
            incrementalUpdate();
        }

        @Override
        protected void clearBounds( Envelope bounds2 ) {
            // ignore because we don't want to clear until first update is received from one of the contained renderers
        }
        
        @Override
        protected void initializeLabelPainter( IRenderContext context2 ) {
            // do nothing
        }

        @Override
        protected void finalizeLabelPainter( IRenderContext context2 ) {
            // do nothing
        }
        
        /**
         * Incrementally updates the display until rendering is complete.
         * 
         * @throws InterruptedException
         * @throws RenderException
         */
        public void incrementalUpdate() throws InterruptedException, RenderException {
        	getExecutor().resetTimeout();
        	final int wait_period = 400;
            
            synchronized (getExecutor()) {
                getExecutor().inUpdate=true;
            }

            CompositeRendererImpl renderer2 = getExecutor().getRenderer();
        	synchronized (getExecutor()) {
        		while( (getMonitor() != null && !getMonitor().isCanceled())
        				&& isRendering(renderer2)
        				&& getExecutor().timeout < TIMEOUT ) {
                    long start=System.currentTimeMillis();
                    getExecutor().wait(wait_period);
        			getExecutor().timeout += start-System.currentTimeMillis();
        			if (isRendering(renderer2)) {
        				renderer2.refreshImage();
        				getExecutor().setState(RENDERING);
        			}
        		}
        	}
        	synchronized (getExecutor()) {
        	    renderer2.refreshImage();
                getExecutor().setState(IRenderer.DONE);
                Graphics2D g = getExecutor().getContext().getImage().createGraphics();
                g.dispose();
        	    getExecutor().inUpdate=false;
            }
        }


        /**
         * @see net.refractions.udig.project.internal.render.impl.RenderJob#postRendering()
         */
        protected void postRendering() {

            if (monitor.isCanceled()) {
                executor.getRenderer().setState(CANCELLED);
                for( RenderExecutor renderer : ((CompositeRendererImpl)executor.getRenderer()).getRenderExecutors() ) {
                    if( renderer.getContext().isVisible() && executor.getState()!=IRenderer.DONE ){
                        renderer.getContext().setStatus(ILayer.WARNING);
                        renderer.getContext().setStatusMessage("Timed out while rendering this layer.  Seems to have blocked, check that the server is up"); //$NON-NLS-1$
                    }
                }
                return;
            }
        }
    }
    
    /**
    *
    * @param renderer2
    * @return
    */
   static boolean isRendering( CompositeRendererImpl renderer2 ) {
       Collection<RenderExecutor> executors = renderer2.getRenderExecutors();
       for( RenderExecutor executor : executors ) {
           if( executor.getContext().isVisible() && (executor.getState()==IRenderer.RENDERING
                   || executor.getState()==IRenderer.STARTING 
                   || executor.getState()==IRenderer.RENDER_REQUEST) )
               return true;
       }
       return false;
   }


    /**
     * Construct <code>CompositeRendererExecutorImpl</code>.
     */
    public RenderExecutorComposite() {
        renderJob = new CompositeRendererJob(this);
    }

    /**
     * @see net.refractions.udig.project.internal.render.impl.RenderExecutorMultiLayer#registerFeatureListener()
     */
    protected void registerFeatureListener() {
        // do nothing
    }

    /**
     * Signals that the timeout for rendering should be reset.
     * <p>
     * The timeout is used to determine whether a renderer has stalled
     * </p>
     */
    protected synchronized void resetTimeout() {
        timeout = 0;
    }

    /**
     * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#getRenderer()
     */
    public CompositeRendererImpl getRenderer() {
        return (CompositeRendererImpl) renderer;
    }

    /**
     * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#setRenderer(net.refractions.udig.project.render.Renderer)
     */
    public void setRenderer( Renderer newRenderer ) {
        renderJob.setSystem(false);
        super.setRendererInternal(newRenderer);
    }

    /**
     * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#stopRendering()
     */
    public void stopRendering() {
        for( RenderExecutor executor : getRenderer().getRenderExecutors() ) {
            executor.stopRendering();
        }
        super.stopRendering();
    }

    protected void resyncState( Renderer renderer ) {

        // FIXME sync sub renderers, witness composite renderer
        //
        // for( Renderer child : parent.children() ) resyncState( child );
    }

    @Override
    protected RendererListener getRendererListener() {
        return new CompositeRendererListener(this);
    }

    @Override
    protected LayerListener getLayerListener() {
        return new LayerListener(this){
            @Override
            public void notifyChanged( Notification msg ) {
                // do nothingsuper.notifyChanged(msg);
            }
        };
    }
    
    @Override
    protected String getRenderJobName() {
        return MessageFormat.format(
                Messages.RenderExecutorImpl_message, new Object[]{getContext().getMap().getName()});  
    }
    
    /**
     * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#visit(net.refractions.udig.project.render.ExecutorVisitor)
     */
    public void visit( ExecutorVisitor visitor ) {
        visitor.visit(this);
    }

    Job refreshJob;
    /**
     * Continually refresh the display until all renderers are done.
     */
    public synchronized void refresh() {
        if (refreshJob == null) {
            refreshJob = new Job(getRenderJobName()){
                @Override
                protected IStatus run( IProgressMonitor monitor ) {
                    try {
                        ((CompositeRendererJob) renderJob).incrementalUpdate();
                    } catch (Exception e) {
                        // log error
                        ProjectPlugin.log(null, e);
                    }
                    return Status.OK_STATUS;
                }

            };
            refreshJob.setSystem(false);
            refreshJob.setPriority(Job.INTERACTIVE);
        }
        refreshJob.schedule();
    }

    @Override
    public synchronized void render() {

        RedrawJob runnable=new RedrawJob();
        runnable.bounds = getRenderBounds();
        if( redraw!=null )
            redraw.interrupt();
        redraw=new Thread(runnable);
        redraw.start();
    }
    
    private volatile Thread redraw ;

    class RedrawJob implements Runnable {

        ReferencedEnvelope bounds;
        
        public void run(  ) {
            
            stopRendering();

            if (!Thread.currentThread().isInterrupted())
                runRenderJob();
            return;
        }
        
        void runRenderJob() {
            dirty = false;
            if (getContext().getMapDisplay() == null || getContext().getMapDisplay().getWidth() < 1
                    || getContext().getMapDisplay().getHeight() < 1)
                return;
            if (RenderExecutorComposite.this.getState() == IRenderer.DISPOSED)
                throw new RuntimeException("attempted to run a disposed renderer"); //$NON-NLS-1$

            clearImage(bounds, RenderExecutorComposite.this);
            renderJob.setName(getRenderJobName());
            renderJob.addRequest(bounds);

            if (!Thread.currentThread().isInterrupted()){
                redraw=null;
            }
        }

    };
}
