package net.refractions.udig.catalog.ui.export;



import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import net.refractions.udig.catalog.ui.CatalogUIPlugin;
import net.refractions.udig.catalog.ui.internal.Messages;
import net.refractions.udig.catalog.ui.workflow.Workflow;
import net.refractions.udig.catalog.ui.workflow.WorkflowWizard;
import net.refractions.udig.catalog.ui.workflow.WorkflowWizardDialog;
import net.refractions.udig.catalog.ui.workflow.WorkflowWizardPage;
import net.refractions.udig.catalog.ui.workflow.Workflow.State;
import net.refractions.udig.ui.PlatformGIS;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;


public class CatalogExport {

    Shell shell;
    private WorkflowWizardDialog dialog;
    protected WorkflowWizard wizard;
    
    static String SHAPEFILE_EXT = ".shp"; //$NON-NLS-1$
    static String POLY_SUFFIX = "_poly"; //$NON-NLS-1$
    static String POINT_SUFFIX = "_point"; //$NON-NLS-1$
    static String LINE_SUFFIX = "_line"; //$NON-NLS-1$

    /**
     * Creates a new instance and calls {@link #init()}
     */
    public CatalogExport() {
        init();
    }

    /**
	 * Creates a new instance and calls {@link #init()} if initialize is true. This method
	 * is here so that initialization can be deferred until later so the
	 * subclass can do other initialization first.  If initialize is false make sure to call
	 * {@link #init()} before using the object.
	 * 
	 * @param initialize
	 *            if true initialize is called
	 */
    public CatalogExport(boolean initialize) {
    	if( initialize ){
    		init();
    	}
    }


    /**
	 * This is a "template method" (see GOF patterns book). It first calls
	 * createWorkFlow() then createPageMapping() then createWorkflowWizard().
	 * It also creates a dialog for the workflow wizard to be opened up in by calling createShell().  
	 * The dialog is blocking.
	 */
	protected final void init() {
        Workflow workflow = createWorkflow();
        Map<Class<? extends State>, WorkflowWizardPage> map = createPageMapping();
        wizard = createWorkflowWizard(workflow,map);

        PlatformGIS.syncInDisplayThread(
            new Runnable() {
                public void run() {
                    shell=createShell();
                }   
            }
        );

        dialog = new WorkflowWizardDialog(
            shell, wizard   
        );

        dialog.setBlockOnOpen(true);
    }   

    /**
     * Must be called in the Display thread.
     */
    Shell createShell() {

        IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();

        if( window!=null ){
            Shell temp = window.getShell();
            if( temp!=null )
                return temp;
        }

        Display d=PlatformUI.getWorkbench().getDisplay();

        if( d==null )
            d=Display.getCurrent();

        return new Shell(d.getActiveShell());

    }

    public WorkflowWizardDialog getDialog() {
        return dialog;
    }

    public void open() {

        Display.getDefault().asyncExec(
            new Runnable() {
                public void run() {
                    dialog.open();  
                };
            }
        );
    }

    public void run(IProgressMonitor monitor, Object context) {

        dialog.getWorkflowWizard().getWorkflow().setContext(context);
        String name=Messages.CatalogExport_taskname; 
        monitor.beginTask(name, 100);
        monitor.setTaskName(name);
        try {
            dialog.runHeadless(new SubProgressMonitor(monitor, 100));
        } finally {
            monitor.done();
        }
    }

    protected Workflow createWorkflow() {

        ExportResourceSelectionState layerState = new ExportResourceSelectionState();
        Workflow workflow = new Workflow(new Workflow.State[]{layerState});
        return workflow;
    }

    protected Map<Class<? extends State>, WorkflowWizardPage> createPageMapping() {
        HashMap<Class<? extends State>, WorkflowWizardPage> map = new HashMap<Class<? extends State>, WorkflowWizardPage>();
        map.put(ExportResourceSelectionState.class, new ExportResourceSelectionPage("Select Layers")); //$NON-NLS-1$

        //TODO: add export support for formats other than shapefile

        return map;
    }

    protected WorkflowWizard createWorkflowWizard(Workflow workflow, Map<Class<? extends State>, WorkflowWizardPage> map) {
        return new CatalogExportWizard(workflow,map);
    }

    public static void setError(final WizardDialog wizardDialog, final String msg, Throwable e) {

        CatalogUIPlugin.log(msg, e);
        
		if( Display.getCurrent()==null ){
			wizardDialog.getShell().getDisplay().asyncExec(new Runnable() {
				public void run() {
			        wizardDialog.setErrorMessage(msg);
				}
			});
		}else{
			wizardDialog.setErrorMessage(msg);
		}
	}
}

