/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.catalog.internal.wms;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IGeoResourceInfo;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.ui.CatalogUIPlugin;
import net.refractions.udig.catalog.ui.ISharedImages;
import net.refractions.udig.catalog.wms.internal.Messages;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Rectangle;
import org.geotools.data.ows.WMSCapabilities;
import org.geotools.data.wms.WebMapServer;
import org.geotools.data.wms.request.GetLegendGraphicRequest;
import org.geotools.data.wms.response.GetLegendGraphicResponse;
import org.geotools.data.wms.xml.WMSSchema;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.ows.ServiceException;
import org.geotools.referencing.CRS;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.NoSuchAuthorityCodeException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Envelope;

/**
 * FeatureType provided by WFS.
 * </p>
 * 
 * @author David Zwiers, Refractions Research
 * @since 0.6
 */
public class WMSGeoResourceImpl extends IGeoResource {

    WMSServiceImpl parent;
    org.geotools.data.ows.Layer layer;
    private IGeoResourceInfo info;
    private ImageDescriptor icon;
    private URL identifier;
    
    private WMSGeoResourceImpl() {
        // should not be used
    }

    /**
     * Construct <code>WMSGeoResourceImpl</code>.
     *
     * @param parent
     * @param layer
     */
    public WMSGeoResourceImpl( WMSServiceImpl parent, org.geotools.data.ows.Layer layer ) {
        this.parent = parent;
        this.layer = layer;
        try {
            identifier=new URL(parent.getIdentifier().toString() + "#" + layer.getName()); //$NON-NLS-1$

        } catch (Throwable e) {
            WmsPlugin.log( null, e);
            identifier = parent.getIdentifier();
        }
    }

    @Override
    public WMSServiceImpl parent( IProgressMonitor monitor ) throws IOException {
        return parent;
    }
    /*
     * @see net.refractions.udig.catalog.IGeoResource#getStatus()
     */
    public Status getStatus() {
        return parent.getStatus();
    }

    public IGeoResourceInfo getInfo( IProgressMonitor monitor ) throws IOException {
        if (info == null) {
            parent.rLock.lock();
            try{
                if (info == null) {
                    info = new WMSResourceInfo( monitor );
                }
            }finally{
                parent.rLock.unlock();
            }
        }
        return info;
    }

    public URL getIdentifier() {
            return identifier;
    }

    /*
     * @see net.refractions.udig.catalog.IGeoResource#resolve(java.lang.Class,
     *      org.eclipse.core.runtime.IProgressMonitor)
     */
    public <T> T resolve( Class<T> adaptee, IProgressMonitor monitor ) throws IOException {
        if (adaptee == null) {
            throw new NullPointerException();
        }
        
//        if (adaptee.isAssignableFrom(IService.class)) {
//            return adaptee.cast( parent);
//        }
        
        if (adaptee.isAssignableFrom(IGeoResource.class)) {
            return adaptee.cast( this );
        }
        
        if (adaptee.isAssignableFrom(IGeoResourceInfo.class)) {
            return adaptee.cast( getInfo(monitor));
        }
        
        if (adaptee.isAssignableFrom(WebMapServer.class)) { 
            return adaptee.cast( parent.getWMS( monitor ));
        }
        
        if (adaptee.isAssignableFrom(org.geotools.data.ows.Layer.class)) {
            return adaptee.cast( layer);
        }
        if (adaptee.isAssignableFrom( ImageDescriptor.class )) {            
            return adaptee.cast( getIcon( monitor ) );
        }
        return super.resolve(adaptee, monitor);
    }
    public IService service( IProgressMonitor monitor ) throws IOException {
        return parent;
    }
    /** Must be the same as resolve( ImageDescriptor.class ) */
    public synchronized ImageDescriptor getIcon( IProgressMonitor monitor ) {
        if( icon == null ) {            
            icon = fetchIcon( monitor );
            if( icon == null ) {
                icon = 
                    CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            }
        }
        return icon;
    }
    /**
     * This method will fetch the Icon associated with this url
     * (if such is available).
     *
     * @param monitor
     * @return Requested Icon or ISharedImages.GRID_OBJ
     */
    protected ImageDescriptor fetchIcon( IProgressMonitor monitor ) {
        try {
            if( monitor != null ) monitor.beginTask( Messages.WMSGeoResourceImpl_acquiring_task , 3 ); 
            if( monitor != null ) monitor.worked(1);
            
            if (layer.getChildren() != null && layer.getChildren().length != 0) {
                //Do not request "parent" layer graphics - this kills Mapserver
                return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            }
            
            WebMapServer wms = parent.getWMS(monitor);
            
            if (wms.getCapabilities().getRequest().getGetLegendGraphic() == null) {
                return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            }
            
            ImageDescriptor imageDescriptor = null;
            try {
                GetLegendGraphicRequest request = wms.createGetLegendGraphicRequest();
                request.setLayer(layer.getName());
                request.setWidth("16"); //$NON-NLS-1$
                request.setHeight("16"); //$NON-NLS-1$
                
                String desiredFormat = null;
                List formats = Arrays.asList(wms.getCapabilities().getRequest().getGetLegendGraphic().getFormatStrings());
                if (formats.contains("image/png")) { //$NON-NLS-1$
                    desiredFormat = "image/png"; //$NON-NLS-1$
                }                    
                if (desiredFormat == null && formats.contains("image/gif")) { //$NON-NLS-1$
                    desiredFormat = "image/gif"; //$NON-NLS-1$
                }                   
                if (desiredFormat != null) {
                    request.setFormat(desiredFormat);
                
                    request.setStyle(""); //$NON-NLS-1$
                    
                    System.out.println(request.getFinalURL().toExternalForm());
                
                    GetLegendGraphicResponse response = wms.issueRequest(request);
     
                    imageDescriptor = ImageDescriptor.createFromImageData(getImageData(response.getInputStream()));
                }
                else {
                    // cannot understand any of the provided formats
                    return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
                }
            } catch (UnsupportedOperationException notAvailable) {
                WmsPlugin.trace( "Icon is not available", notAvailable ); //$NON-NLS-1$                
                return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            } catch (ServiceException e) {
                WmsPlugin.trace( "Icon is not available", e ); //$NON-NLS-1$                
                return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            } catch (SWTException exc){
                WmsPlugin.trace( "Icon is not available or has unsupported format", exc ); //$NON-NLS-1$                
                return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            }

            Image image = null;
            Image swatch = null;
            try {
                if( monitor != null ) monitor.worked(2);
                if( monitor != null ) monitor.subTask(Messages.WMSGeoResourceImpl_downloading_icon); 
                image = imageDescriptor.createImage();
                Rectangle bound = image.getBounds();
                if( bound.width == 16 && bound.height == 16 ) {
                    final ImageData data = (ImageData) image.getImageData().clone();
                    return new ImageDescriptor(){
                        public ImageData getImageData() {
                            return (ImageData) data.clone(); 
                        }                                
                    };
                }                                                      
                if( bound.height < 16 || bound.width < 16){
                    if( WmsPlugin.getDefault().isDebugging() ) System.out.println( "Icon scaled up to requested size"); //$NON-NLS-1$                                        
                    final ImageData data = image.getImageData().scaledTo( 16, 16 );
                    return new ImageDescriptor(){
                        public ImageData getImageData() {
                            return (ImageData) data.clone(); 
                        }                                
                    };                       
                }
                swatch = new Image( null, 16, 16 );
                GC gc = new GC( swatch );
                int sy = 0; //(bound.height / 2 ) - 8;
                int sx = 0;
                int sw = 0;
                int sh = 0;
                ImageData contents = image.getImageData();
                if( contents == null ){
                    return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_MISSING );
                }
                if( contents.maskData != null ){
                    // ((width + 7) / 8 + (maskPad - 1)) / maskPad * maskPad
                    int maskPad = contents.maskPad;
                    int scanLine = ((contents.width+7)/8 + (maskPad -1 )) / maskPad * maskPad;
                    // skip leading mask ...
                    SKIPY: for( int y = 0; y < contents.height/2; y++ ){
                        sy = y;
                        for( int x = 0; x < contents.width/2; x++ ){
                            int mask = contents.maskData[ y*scanLine + x ]; 
                            if( mask != 0) break SKIPY;
                        }                        
                    }
                    SKIPX: for( int x = 0; x < contents.width/2; x++ ){
                        sx = x;
                        for( int y = sy; y < contents.height/2; y++ ){
                            int mask = contents.maskData[ y*scanLine + x ]; 
                            if( mask  != 0) break SKIPX;
                        }                        
                    }
                    sh = Math.min( contents.height - sy, 16 );
                    sw = Math.min( contents.width - sx, 16 );
                    if( WmsPlugin.getDefault().isDebugging() ) System.out.println( "Mask offset to "+sx+"x"+sy); //$NON-NLS-1$ //$NON-NLS-2$                        
                }
                else if( contents.alphaData != null ){
                    SKIPY: for( int y = 0; y < contents.height/2; y++ ){
                        sy = y;
                        for( int x = 0; x < contents.width/2; x++ ){
                            int alpha = contents.alphaData[ y*contents.width + x ]; 
                            if( alpha != 0) break SKIPY;
                        }                        
                    }
                    SKIPX: for( int x = 0; x < contents.width/2; x++ ){
                        sx = x;
                        for( int y = sy; y < contents.height/2; y++ ){
                            int alpha = contents.alphaData[ y*contents.width + x ]; 
                            if( alpha != 0) break SKIPX;
                        }
                    }
                    sh = Math.min( contents.height - sy, 16 );
                    sw = Math.min( contents.width - sx, 16 );                    
                    if( WmsPlugin.getDefault().isDebugging() ) System.out.println( "Alpha offset to "+sx+"x"+sy); //$NON-NLS-1$ //$NON-NLS-2$                        
                }
                else {
                    // try ignoring "white"
                    int depth = contents.depth;
                    int scanLine = contents.bytesPerLine;                                                            
                    SKIPY: for( int y = 0; y < contents.height/2; y++ ){
                        sy = y;
                        for( int x = 0; x < contents.width/2; x++ ){
                            int datum = contents.data[ y*scanLine + x*depth ]; 
                            if( datum != 0) break SKIPY;
                        }                        
                    }
                    SKIPX: for( int x = 0; x < contents.width/2; x++ ){
                        sx = x;
                        for( int y = sy; y < contents.height/2; y++ ){
                            int datum = contents.data[ y*scanLine + x*depth ]; 
                            if( datum != 0) break SKIPX;
                        }                        
                    }
                    sh = Math.min( contents.height - sy, 16 );
                    sw = Math.min( contents.width - sx, 16 );                    
                    if( WmsPlugin.getDefault().isDebugging() ) System.out.println( "Alpha offset to "+sx+"x"+sy); //$NON-NLS-1$ //$NON-NLS-2$                        
                }
//                else {
//                    sh = Math.min( bound.height, bound.width );
//                    sw = Math.min( bound.height, bound.width );
//                }
                if( WmsPlugin.getDefault().isDebugging() ) System.out.println( "Image resized to "+sh+"x"+sw); //$NON-NLS-1$ //$NON-NLS-2$
                
                gc.drawImage( image, sx, sy, sw, sh, 0, 0, 16, 16 );
                final ImageData data = (ImageData) swatch.getImageData().clone();
                return new ImageDescriptor(){
                    public ImageData getImageData() {
                        return (ImageData) data.clone(); 
                    }                                
                };
                //
            }
            finally {
                if( image != null ) {
                    image.dispose();
                }
                if( swatch != null ) {
                    swatch.dispose();
                }
                if( monitor != null ) monitor.worked(3);
            }
        }
        catch (IOException t) {
            WmsPlugin.trace( "Could not get icon", t ); //$NON-NLS-1$
            return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_MISSING );
        }
    }
    
    private ImageData getImageData(InputStream in) {
        ImageData result = null;
        if (in != null) {
            try {
                result = new ImageData(in);
            } catch (SWTException e) {
                if (e.code != SWT.ERROR_INVALID_IMAGE)
                    throw e;
                // fall through otherwise
            } finally {
                try {
                    in.close();
                } catch (IOException e) {
                    //System.err.println(getClass().getName()+".getImageData(): "+
                    //  "Exception while closing InputStream : "+e);
                }
            }
        }
        return result;
    }
    
    /*
     * @see net.refractions.udig.catalog.IResolve#canResolve(java.lang.Class)
     */
    public <T> boolean canResolve( Class<T> adaptee ) {
        if (adaptee == null) {
            return false;
        }

        if (adaptee.isAssignableFrom(IGeoResource.class)
                || adaptee.isAssignableFrom(WebMapServer.class)
                || adaptee.isAssignableFrom(org.geotools.data.ows.Layer.class)
                || adaptee.isAssignableFrom(ImageDescriptor.class)
                || adaptee.isAssignableFrom(IService.class)||
                super.canResolve(adaptee)) {
            return true;
        }

        return false;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getMessage()
     */
    public Throwable getMessage() {
        return parent.getMessage();
    }

    private class WMSResourceInfo extends IGeoResourceInfo {
        WMSResourceInfo( IProgressMonitor monitor ) throws IOException {
            WebMapServer wms = parent.getWMS( monitor );
            WMSCapabilities caps = wms.getCapabilities();
            
            org.opengis.geometry.Envelope env = null;
            CoordinateReferenceSystem crs = null;
            try {
                crs = CRS.decode("EPSG:4326"); //$NON-NLS-1$
                env = parent.getWMS(null).getEnvelope(layer, crs);                
            } catch (NoSuchAuthorityCodeException e) {
                throw (IOException) new IOException( Messages.WMSGeoResourceImpl_bounds_unavailable).initCause( e ); 
            } catch (FactoryException e) {
                throw (IOException) new IOException( Messages.WMSGeoResourceImpl_bounds_unavailable).initCause( e ); 
			}            
            bounds = new ReferencedEnvelope(new Envelope(env.getMinimum(0), env.getMaximum(0), 
                    env.getMinimum(1), env.getMaximum(1)), crs );
            
            String parentid = parent != null && parent.getIdentifier() != null ? getIdentifier().toString() : ""; //$NON-NLS-1$
            name = layer.getName();            
            List<String> keywordsFromWMS = new ArrayList<String>();            
            if (caps.getService().getKeywordList() != null) {
                keywordsFromWMS.addAll( Arrays.asList( caps.getService().getKeywordList() ));
            }
                        
            if (layer.getKeywords() != null) {
                keywordsFromWMS.addAll( Arrays.asList( layer.getKeywords() ) );           
            }            
            keywordsFromWMS.add("WMS"); //$NON-NLS-1$
            keywordsFromWMS.add( layer.getName() );
            keywordsFromWMS.add( caps.getService().getName() );
            keywordsFromWMS.add( parentid );
            keywords = keywordsFromWMS.toArray(new String[keywordsFromWMS.size()]);
                        
            if (layer.get_abstract() != null && layer.get_abstract().length() != 0) {
                description = layer.get_abstract();
            } else {
                description = caps.getService().get_abstract();
            }            
            description = caps.getService().get_abstract();
            
            if (layer.getTitle() != null && layer.getTitle().length() != 0) {
                title = layer.getTitle();
            } else {
                title = caps.getService().getTitle();
            }            
            super.icon = CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.GRID_OBJ );
            // icon = fetchIcon( monitor );            
        }                
        public String getName() {
            return name;
        }
        public URI getSchema() {
            return WMSSchema.NAMESPACE;
        }
        public String getTitle() {
            return title;
        }
    }
}