/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.core;

import java.io.IOException;

import net.refractions.udig.core.internal.CorePlugin;

import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;

/**
 * Methods to help manage adapters
 * @author jones
 * @since 1.1.0
 */
public class AdapterUtil {
    public static final AdapterUtil instance=new AdapterUtil();
    
    private AdapterUtil(){}
    
    /**
     * Determines if a class can be adapted based on its string representation.  
     *
     * @param targetClass class name to check if adaptation can go to
     * @param obj source object to adapt from
     * @return
     */
    public boolean canAdaptTo(String targetClass, Object obj) {
        ClassLoader classLoader = obj.getClass().getClassLoader();
        return canAdaptTo(targetClass, obj, classLoader);
    }
    
    /**
     * Determines if a class can be adapted based on its string representation.  
     *
     * @param targetClass class name to check if adaptation can go to
     * @param obj source object to adapt from
     * @param classLoader object to instantiate the class with
     * @return
     */
    public boolean canAdaptTo(String targetClass, Object obj, ClassLoader classLoader) {
        Class< ? > target=null;
        if (classLoader == null) {
            classLoader = obj.getClass().getClassLoader();
        }
        try {
            target = classLoader.loadClass(targetClass);
        } catch (Throwable e) {
            //do nothing this is ok.
            return false;
        }
        if( target.isAssignableFrom(obj.getClass()) )
            return true;
        
        //see if platform can adapt the object
        if (Platform.getAdapterManager().hasAdapter(obj, targetClass) )
            return true;
        
        // see if the object is a blocking adaptable object and can adapt to the correct class type
        IBlockingAdaptable blockingAdaptable=getBlockingAdapter(obj);
        if( blockingAdaptable!=null && blockingAdaptable.canAdaptTo(target))
            return true;
        
        // see if the object is adaptable and can adapt to the class type.
        IAdaptable adaptable=getAdaptable(obj);
        if( adaptable!=null ){
            return adaptable.getAdapter(target)!=null;
        }
        return false;
    }
    
    public IAdaptable getAdaptable(Object obj) {
        if( obj instanceof IAdaptable ){
            return (IAdaptable) obj;
        }
        if (Platform.getAdapterManager().hasAdapter(obj, IAdaptable.class.getSimpleName()) )
            return (IAdaptable) Platform.getAdapterManager().getAdapter(obj, IAdaptable.class);
        return null;
    }

    public IBlockingAdaptable getBlockingAdapter(Object obj) {
        if( obj instanceof IBlockingAdaptable ){
            return (IBlockingAdaptable) obj;
        }
        if (Platform.getAdapterManager().hasAdapter(obj, IBlockingAdaptable.class.getName()) )
            return (IBlockingAdaptable) Platform.getAdapterManager().getAdapter(obj, IBlockingAdaptable.class);
        if (obj instanceof IAdaptable) {
            IAdaptable adapter = (IAdaptable) obj;
            return (IBlockingAdaptable) adapter.getAdapter(IBlockingAdaptable.class);
        }
        return null;
    }

    /**
     * Since the target object may not be the object that the operation actually operates on, the
     * getOperationTarget() finds the real object and returns it or null if for some reason the
     * operation can be performed on the target.
     * <p>
     * Example:
     * </p>
     * A FeatureType readonly operation is called on a IResolve. The getOperationTarget would
     * resolve(FeatureSource.class) and return the Feature Source for the operation.
     * 
     * @param target the object the action is called on.
     * @param definition the Configuration element definition of the operation.
     * @param monitor The progress monitor for the operation job.
     * @return
     * @throws IOException 
     */
    @SuppressWarnings("unchecked")
    public <T> T getOperationTarget( String targetClass, Object obj, IProgressMonitor monitor ) throws IOException{
        Class< T > target=null;
        try {
            target=(Class<T>) obj.getClass().getClassLoader().loadClass(targetClass);
            return getOperationTarget(target, obj, monitor);
        } catch (ClassNotFoundException e) {
            //do nothing.
            CorePlugin.log( "This exception is bad.  The framework should not allow this to be reached.",e); //$NON-NLS-1$
            return null;
        }
        
        
         
    }

    /**
     * Since the target object may not be the object that the operation actually operates on, the
     * getOperationTarget() finds the real object and returns it or null if for some reason the
     * operation can be performed on the target.
     * <p>
     * Example:
     * </p>
     * A FeatureType readonly operation is called on a IResolve. The getOperationTarget would
     * resolve(FeatureSource.class) and return the Feature Source for the operation.
     * 
     * @param target the object the action is called on.
     * @param definition the Configuration element definition of the operation.
     * @param monitor The progress monitor for the operation job.
     * @return
     * @throws IOException 
     */
    public <T> T getOperationTarget( Class<T> target, Object obj, IProgressMonitor monitor ) throws IOException{
        
            if( target.isAssignableFrom(obj.getClass()) )
                return target.cast(obj);
        if (Platform.getAdapterManager().hasAdapter(obj, target.getName()) ){
            return target.cast(Platform.getAdapterManager().loadAdapter(obj, target.getName()));
        }
        IBlockingAdaptable blockingAdaptable=getBlockingAdapter(obj);
        if( blockingAdaptable!=null && blockingAdaptable.canAdaptTo(target))
            return blockingAdaptable.getAdapter(target, monitor);
        if (obj instanceof IAdaptable) {
            IAdaptable adaptable = (IAdaptable) obj;
            return target.cast(adaptable.getAdapter(target));
        }
        return null;
    }

}
