package net.refractions.udig.catalog.internal.ui;

import java.io.IOException;
import java.util.Iterator;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.ui.CatalogTreeViewer;
import net.refractions.udig.catalog.ui.CatalogUIPlugin;
import net.refractions.udig.catalog.ui.StatusLineMessageBoardAdapter;
import net.refractions.udig.catalog.ui.internal.Messages;
import net.refractions.udig.internal.ui.IDropTargetProvider;
import net.refractions.udig.internal.ui.UiPlugin;
import net.refractions.udig.ui.ProgressManager;
import net.refractions.udig.ui.UDIGDragDropUtilities;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.GroupMarker;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.part.ISetSelectionTarget;
import org.eclipse.ui.part.ViewPart;
import org.osgi.service.prefs.BackingStoreException;

/**
 * Catalog view for visualization and management if resources.
 * <p>
 * This class will be rather heavy on documentation, because it central (literally) to the uDig
 * applicaiton, and because it is one of the first views we are creating.
 * </p>
 * <p>
 * Of Note:
 * <ul>
 * <li>The catalog is strange in that there is only *one*, represent a global registry of all the
 * data sources in use by any uDig plugins.
 * <li>
 * </ul>
 * </p>
 * From the requirements document (where CatalogView is known as LocalCatalog): <i>The Local Catalog
 * serves as a central repository of data and server information.
 * <p>
 * Non-Functional Requirements:
 * <ul>
 * <li><b>Ease of Data Location </b>, intent is for the user to be separated from the data source;
 * so they need as little technical knowledge as possible.
 * <li><b>Security </b>, name/password should be left out of the export/share.
 * </ul>
 * </p>
 * <p>
 * Functional Requirements:
 * <ul>
 * <li><b>Servers </b>, store server connection information for sharing between projects.
 * <li><b>Data Directories </b>, store data directories for sharing between projects.
 * <li><b>Metadata </b>, provide access to metadata on Servers/Data Directories.
 * <li><b>Data Discovery </b>, provide enough information for a user to define a new layer in their
 * context.
 * <li><b>Persist Settings </b>, permit exporting and sharing DataStore connection information.
 * <li><b>DataStores Management </b>, lookup actualized DataStores that are in use.
 * <li><b>Missing Data </b>, entries referred to by imported projects should be maintained,
 * allowing the user one location to correct data connection information.
 * </ul>
 * </p>
 * <p>
 * Catalog View has a strong interaction with the preferences maintained by the local installation
 * of uDig for the current user.
 * </p>
 */
public class CatalogView extends ViewPart implements ISetSelectionTarget, IDropTargetProvider {

    /** <code>VIEW_ID</code> field */
    public static final String VIEW_ID = "net.refractions.udig.catalog.ui.CatalogView"; //$NON-NLS-1$

    CatalogTreeViewer treeviewer;

    Action removeAction; // addAction

    private Action saveAction;
    private Action loadAction;

    /**
     * Creates the SWT controls for this workbench part.
     * <p>
     * The details (from IWorkbenchPart.createPartControl( Composite ))
     * </p>
     * <p>
     * Multi-step process:
     * <ol>
     * <li>Create one or more controls within the parent.</li>
     * <li>Set the parent layout as needed.</li>
     * <li>Register any global actions with the <code>IActionService</code>. </li>
     * <li>Register any popup menus with the <code>IActionService</code>. </li>
     * <li>Register a selection provider with the <code>ISelectionService</code> (optional).</li>
     * </ol>
     * </p>
     * 
     * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
     * @param parent
     */
    public void createPartControl( Composite parent ) {
        // create viewer
//        treeviewer = new CatalogTreeViewer(parent, false);
        treeviewer = new CatalogTreeViewer(parent, true);
        treeviewer.setMessageBoard(new StatusLineMessageBoardAdapter(getViewSite().getActionBars().getStatusLineManager()));

        UDIGDragDropUtilities.addDragDropSupport(treeviewer, this);

        getSite().setSelectionProvider(treeviewer);
        // Create menu and toolbars
        createActions();
        createMenu();
        createToolbar();
        createContextMenu();
        // hookGlobalActions();

        // restore state (from previous session)

    }

    private void createContextMenu() {
        final MenuManager contextMenu = new MenuManager();

        contextMenu.setRemoveAllWhenShown(true);
        contextMenu.addMenuListener(new IMenuListener(){

            public void menuAboutToShow( IMenuManager mgr ) {
                contextMenu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));
                contextMenu.add(new Separator());
                contextMenu.add(removeAction);
                IWorkbenchWindow window = getSite().getWorkbenchWindow();
                IAction action = ActionFactory.IMPORT.create(window);
                contextMenu.add(action);
                contextMenu.add(new Separator());
                contextMenu.add(UiPlugin.getDefault().getOperationMenuFactory().getContextMenu(treeviewer.getSelection()));
                contextMenu.add(new Separator());
                contextMenu.add(ActionFactory.EXPORT.create(getSite().getWorkbenchWindow()));
            }

        });

        // Create menu.
        Menu menu = contextMenu.createContextMenu(treeviewer.getControl());
        treeviewer.getControl().setMenu(menu);

        // Register menu for extension.
        getSite().registerContextMenu(contextMenu, treeviewer);

    }

    private void createActions() {
        removeAction = new Action(){
            public void run() {
                freeDatasource();
            }
        };
        
        Messages.initAction(removeAction, "action_remove"); //$NON-NLS-1$
        removeAction.setEnabled(false);
        removeAction.setImageDescriptor(Images.getDescriptor(ImageConstants.REMOVE_CO));
        removeAction.setActionDefinitionId("org.eclipse.ui.edit.delete"); //$NON-NLS-1$
        getViewSite().getActionBars().setGlobalActionHandler(ActionFactory.DELETE.getId(), removeAction);
        getSite().getKeyBindingService().registerAction(removeAction);
        
        PlatformUI.getWorkbench().getHelpSystem().setHelp(removeAction,
                IHelpContextIds.REMOVE_SERVICE_ACTION);

        saveAction = new Action(Messages.CatalogView_save_label){ 
            public void run() {
                try {
                    CatalogPlugin.getDefault().storeToPreferences(ProgressManager.instance().get());
                } catch (BackingStoreException e) {
                    CatalogPlugin.log(null, e);
                } catch (IOException e) {
                    CatalogPlugin.log(null, e);
                }
            }
        };

        loadAction = new Action(Messages.CatalogView_load_label){ 
            public void run() {
                try {
                    CatalogPlugin.getDefault().restoreFromPreferences();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        };

        // Add selection listener.
        treeviewer.addSelectionChangedListener(new ISelectionChangedListener(){
            public void selectionChanged( SelectionChangedEvent event ) {
                updateActionEnablement();
            }
        });
    }

    void registerDatasource() {
        // Call to wizard here...
    }

    void freeDatasource() {
        // Free selected data source - but only if it is not
        // in use...

        IStructuredSelection sel = (IStructuredSelection) treeviewer.getSelection();
        for( Iterator iter = sel.iterator(); iter.hasNext(); ) {
            @SuppressWarnings("all")//$NON-NLS-1$
            Object o = iter.next();
            if (o instanceof IService)
                remove((IService) o);
            else if (o instanceof IGeoResource)
                remove((IGeoResource) o);
        }
    }

    private void remove( IService service ) {
        CatalogPlugin.getDefault().getCatalogs()[0].remove(service);
    }

    private void remove( IGeoResource service ) {
        try {
            remove(service.service(null));
        } catch (IOException e) {
            CatalogUIPlugin.log(null, e);
        }
    }

    void updateActionEnablement() {
        IStructuredSelection sel = (IStructuredSelection) treeviewer.getSelection();
        removeAction.setEnabled(sel.size() > 0);
    }

    /**
     * Create menu with refresh option.
     */
    private void createMenu() {
        IMenuManager mgr = getViewSite().getActionBars().getMenuManager();
        mgr.add(saveAction);
        mgr.add(loadAction);
    }

    /**
     * Create toolbar with new and delete buttons.
     */
    private void createToolbar() {
        IToolBarManager mgr = getViewSite().getActionBars().getToolBarManager();
        // mgr.add(addAction);

        IWorkbenchWindow window = getSite().getWorkbenchWindow();
        IAction action = ActionFactory.IMPORT.create(window);

        action.setImageDescriptor(Images
                .getDescriptor(ImageConstants.PATH_ETOOL + "import_wiz.gif")); //$NON-NLS-1$
        mgr.add(action);

        mgr.add(removeAction);
    }

    /**
     * Asks this view take focus within the workbench.
     * <p>
     * From IWorkbenchPart: Clients should not call this method (the workbench calls this method at
     * appropriate times). To have the workbench activate a part, use
     * <code>IWorkbenchPage.activate(IWorkbenchPart) instead</code>.
     * </p>
     * <p>
     * Used to set the focus to the appropriate control, for us that is the treeviewer. But if we
     * were smart we could send the user off to a search field or something they actually need (like
     * a broken datastore) based on context.
     * </p>
     */
    public void setFocus() {
        treeviewer.getControl().setFocus();
    }
    /**
     * @return Returns the treeviewer.
     */
    public CatalogTreeViewer getTreeviewer() {
        return treeviewer;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.eclipse.ui.part.ISetSelectionTarget#selectReveal(org.eclipse.jface.viewers.ISelection)
     */
    public void selectReveal( ISelection selection ) {
        treeviewer.setSelection(selection, true);
    }

    public Object getTarget(DropTargetEvent event) {
        return this;
    }
}
