/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.project.ui;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.internal.ui.UiPlugin;
import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.IMap;
import net.refractions.udig.project.IProject;
import net.refractions.udig.project.IProjectElement;
import net.refractions.udig.project.interceptor.ShowViewInterceptor;
import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.project.internal.Map;
import net.refractions.udig.project.internal.Project;
import net.refractions.udig.project.internal.ProjectElement;
import net.refractions.udig.project.internal.ProjectPlugin;
import net.refractions.udig.project.internal.commands.AddLayersCommand;
import net.refractions.udig.project.internal.commands.CreateMapCommand;
import net.refractions.udig.project.internal.impl.ProjectRegistryImpl;
import net.refractions.udig.project.internal.render.CompositeRenderContext;
import net.refractions.udig.project.internal.render.RenderContext;
import net.refractions.udig.project.internal.render.RenderFactory;
import net.refractions.udig.project.internal.render.RenderManager;
import net.refractions.udig.project.internal.render.Renderer;
import net.refractions.udig.project.internal.render.RendererCreator;
import net.refractions.udig.project.internal.render.SelectionLayer;
import net.refractions.udig.project.internal.render.ViewportModel;
import net.refractions.udig.project.internal.render.impl.RenderContextImpl;
import net.refractions.udig.project.internal.render.impl.RendererCreatorImpl;
import net.refractions.udig.project.render.RenderException;
import net.refractions.udig.project.render.displayAdapter.IMapDisplay;
import net.refractions.udig.project.render.displayAdapter.MapDisplayEvent;
import net.refractions.udig.project.ui.commands.OpenProjectElementCommand;
import net.refractions.udig.project.ui.internal.ApplicationGISInternal;
import net.refractions.udig.project.ui.internal.MapEditor;
import net.refractions.udig.project.ui.internal.Messages;
import net.refractions.udig.project.ui.internal.ProjectUIPlugin;
import net.refractions.udig.project.ui.internal.UDIGEditorInputDescriptor;
import net.refractions.udig.project.ui.internal.tool.ToolContext;
import net.refractions.udig.project.ui.internal.tool.display.ToolManager;
import net.refractions.udig.project.ui.internal.tool.impl.ToolContextImpl;
import net.refractions.udig.project.ui.tool.IToolContext;
import net.refractions.udig.project.ui.tool.IToolManager;
import net.refractions.udig.ui.PlatformGIS;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.ISafeRunnable;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IViewReference;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.opengis.filter.Filter;

/**
 * A facade into udig to simplify operations such as getting the active map and
 * openning a map editor.
 * 
 * @author jeichar
 * @since 0.9.0
 */
public class ApplicationGIS {

	private static java.util.Map<Class, List<UDIGEditorInputDescriptor>> editorInputDescriptors = new HashMap<Class, List<UDIGEditorInputDescriptor>>();
	private static IToolManager toolManager;

	/**
	 * Obtains the current project.
	 * 
	 * @return The current active project
	 */
	public static IProject getActiveProject() {
		Project project = ProjectPlugin.getPlugin().getProjectRegistry()
				.getCurrentProject();

		if (project != null)
			return project;

		return ProjectPlugin.getPlugin().getProjectRegistry()
				.getDefaultProject();
	}

	/**
	 * Return all Projects. The list is unmodifiable.
	 * 
	 * @return all Projects.
	 */
	public static List<? extends IProject> getProjects() {
		return Collections.unmodifiableList(ProjectPlugin.getPlugin()
				.getProjectRegistry().getProjects());
	}

	/**
	 * May return null if the active editor is not a Map Editor.
	 * 
	 * @return the map contained by the current MapEditor or null if the active
	 *         editor is not a map editor.
	 */
	public static IMap getActiveMap() {
		try {
			final IEditorPart[] editor = new IEditorPart[1];
			if (Display.getCurrent() != null) {
				// method being called from ui thread, go ahead
				try {
					editor[0] = PlatformUI.getWorkbench()
							.getActiveWorkbenchWindow().getActivePage()
							.getActiveEditor();
				} catch (Throwable t) {
					// do nothing
				}
			} else {
				// need to be in ui thread
				PlatformGIS.syncInDisplayThread(new Runnable() {
					public void run() {
						try {
							editor[0] = PlatformUI.getWorkbench()
									.getActiveWorkbenchWindow().getActivePage()
									.getActiveEditor();
						} catch (Throwable t) {
							// do nothing
						}
					}
				});
			}

			if (editor[0] instanceof MapEditor) {
				return ((MapEditor) editor[0]).getMap();
			}
		} catch (Exception e) {
			return null;
		}

		return null;
	}

	/**
	 * Returns all open maps. May return null if no Map Editors exist.
	 * 
	 * @return a list of maps contained.
	 */
	public static List<? extends IMap> getOpenMaps() {
		try {

			// For some reason, getting the active workbench doesn't seem to
			// work??!! So looping
			// through all the workbenches and all their pages was the slow way
			// to go...
			// IEditorReference[] editors =
			// PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getEditorReferences();
			IWorkbenchWindow[] wWindows = PlatformUI.getWorkbench()
					.getWorkbenchWindows();

			List<IMap> maps = new ArrayList<IMap>();
			for (IWorkbenchWindow wWindow : wWindows) {
				IWorkbenchPage[] wPages = wWindow.getPages();
				for (IWorkbenchPage wPage : wPages) {
					IEditorReference[] editors = wPage.getEditorReferences();
					for (IEditorReference editor : editors) {
						IEditorPart editor2 = editor.getEditor(false);
						if (editor2 instanceof MapEditor) {
							maps.add(((MapEditor) editor2).getMap());
						}
					}
				}
			}
			if (maps.size() == 0)
				return maps;
			return maps;
		} catch (NullPointerException e) {
			return null;
		}
	}

	/**
	 * Opens a Map editor for the provided map, This is a non-blocking call.
	 * Equivalent to openMap(map, false);
	 * 
	 * @param map
	 *            the map to open. Must be an instance of Map.
	 */
	public static void openMap(IMap map) {
		openMap(map, false);
	}

	/**
	 * Opens a Map editor for the provided map.
	 * 
	 * @param map
	 *            the map to open. Must be an instance of Map.
	 * @param wait
	 *            indicates whether to wait for the map to open before
	 *            returning.
	 */
	public static void openMap(IMap map, boolean wait) {
		openProjectElement(map, wait);
	}

	/**
	 * creates a map and opens an editor for the map.
	 * 
	 * @param a
	 *            list of IGeoResources. Each resource will be a layer in the
	 *            created map.
	 */
	public static void createAndOpenMap(List<IGeoResource> resources) {
		CreateMapCommand command = new CreateMapCommand(null, resources, null);
		getActiveProject().sendSync(command);
		openMap(command.getCreatedMap());
	}

	/**
	 * creates a map and opens an editor for the map.
	 * 
	 * @param a
	 *            list of IGeoResources. Each resource will be a layer in the
	 *            created map.
	 * @param owner
	 *            the project that will contain the map. owner must be an
	 *            instance of Project. If it is obtained using the framework
	 *            then this will always be the case.
	 */
	public static void createAndOpenMap(List<IGeoResource> resources,
			IProject owner) {
		CreateMapCommand command = new CreateMapCommand(null, resources, owner);
		getActiveProject().sendSync(command);
		openMap(command.getCreatedMap());
	}

	/**
	 * creates a map and opens an editor for the map.
	 * 
	 * @param a
	 *            list of IGeoResources. Each resource will be a layer in the
	 *            created map.
	 * @param owner
	 *            the project that will contain the map. owner must be an
	 *            instance of Project. If it is obtained using the framework
	 *            then this will always be the case.
	 * @param wait
	 *            indicates whether to wait for the map to open before
	 *            returning.
	 */
	public static void createAndOpenMap(List<IGeoResource> resources,
			IProject owner, boolean wait) {
		CreateMapCommand command = new CreateMapCommand(null, resources, owner);
		getActiveProject().sendSync(command);
		openMap(command.getCreatedMap(), wait);
	}

	/**
	 * Gets a reference to a view. If the view has not been opened previously
	 * then the view will be opened.
	 * 
	 * @param show
	 *            whether to show the view or not.
	 * @param id
	 *            the id of the view to show.
	 * @return returns the view or null if the view does not exist
	 */
	public static IViewPart getView(boolean show, String id) {
		IWorkbenchPage page = PlatformUI.getWorkbench()
				.getActiveWorkbenchWindow().getActivePage();
		IViewReference[] view = page.getViewReferences();
		IViewReference infoRef = null;
		for (IViewReference reference : view) {
			if (reference.getId().equals(id)) {
				infoRef = reference;
				break;
			}
		}
		// JONES: need to get the part and set the selection to null so that the
		// last selected
		// feature will not flash (because it will not be in list any more).
		IViewPart infoView = null;
		if (infoRef == null) {
			try {
				infoView = page.showView(id);
			} catch (PartInitException e1) {
				return null;
			}
			if (infoView == null) {
				return null;
			}
		}
		if (infoRef != null)
			return (IViewPart) infoRef.getPart(show);

		return null;

	}

	/**
	 * Runs the given runnable in a protected mode. Exceptions thrown in the
	 * runnable are logged and passed to the runnable's exception handler. Such
	 * exceptions are not rethrown by this method.
	 */
	public static void run(ISafeRunnable request) {
		PlatformGIS.run(request);
	}

	/**
	 * Returns an editor input for the type passed in. Processes the
	 * editorInputs extension point.
	 * 
	 * @param type
	 * @return an editor input for the type passed in.
	 */
	public static List<UDIGEditorInputDescriptor> getEditorInputs(
			final Class<? extends IProjectElement> type) {
		if (!editorInputDescriptors.containsKey(type)) {
			synchronized (editorInputDescriptors) {
				if (editorInputDescriptors.containsKey(type))
					return editorInputDescriptors.get(type);

				final List<UDIGEditorInputDescriptor> newInputs = new ArrayList<UDIGEditorInputDescriptor>();

				IExtensionRegistry registry = Platform.getExtensionRegistry();
				IExtensionPoint extensionPoint = registry
						.getExtensionPoint("net.refractions.udig.project.ui.editorInputs"); //$NON-NLS-1$
				IExtension[] extensions = extensionPoint.getExtensions();

				for (int i = 0; i < extensions.length; i++) {
					IConfigurationElement[] elements = extensions[i]
							.getConfigurationElements();

					for (int j = 0; j < elements.length; j++) {
						try {
							Object object = elements[j]
									.createExecutableExtension("projectElement"); //$NON-NLS-1$
							if (object.getClass().isAssignableFrom(type)) {
								UDIGEditorInputDescriptor input = new UDIGEditorInputDescriptor();
								input.setEditorID(elements[j]
										.getAttribute("editorPartID")); //$NON-NLS-1$
								input.setName(elements[j].getAttribute("name")); //$NON-NLS-1$
								input.setExtensionElement(elements[j]);
								input.setType(type);
								newInputs.add(input);
							}
						} catch (CoreException e) {
							e.printStackTrace();
						}
					}
				}

				editorInputDescriptors.put(type, newInputs);
			}
		}

		return editorInputDescriptors.get(type);
	}

	/**
	 * Returns the ToolManager singleton.
	 * 
	 * @return the ToolManager singleton.
	 */
	public static IToolManager getToolManager() {
		synchronized (ToolManager.class) {
			if (toolManager == null) {

				String prefConstant = IToolManager.P_TOOL_MANAGER;
				String xpid = IToolManager.XPID;
				String idField = IToolManager.ATTR_ID;
				String classField = IToolManager.ATTR_CLASS;

				IToolManager result = (IToolManager) UiPlugin
						.lookupConfigurationObject(IToolManager.class,
								ProjectUIPlugin.getDefault()
										.getPreferenceStore(),
								ProjectUIPlugin.ID, prefConstant, xpid,
								idField, classField);
				if (result != null) {
					toolManager = result;
				} else {
					toolManager = new ToolManager();
				}
			}
		}
		return toolManager;
	}

	/**
	 * Returns the IEditorInput instance that wraps the element argument.
	 * 
	 * @return the IEditorInput instance that wraps the element argument.
	 */
	public static UDIGEditorInput getInput(IProjectElement element) {
		List<UDIGEditorInputDescriptor> descriptors = getEditorInputs(element
				.getClass());
		for (UDIGEditorInputDescriptor descriptor : descriptors) {
			UDIGEditorInput input = descriptor.createInput(element);
			if (input != null) {
				return input;
			}
		}
		return null;
	}

	/**
	 * Creates a Tools Context out of Map.
	 * 
	 * @param map
	 *            that the context interacts with
	 * @return a ToolContext
	 * @see ToolContext
	 */
	public static IToolContext createContext(IMap map) {
		if (map instanceof net.refractions.udig.project.internal.Map) {
			ToolContext context = new ToolContextImpl();
			context
					.setMapInternal((net.refractions.udig.project.internal.Map) map);
			return context;
		}
		return null;
	}

	/**
	 * Make layers from the resourceList and adds the layers to the map.
	 * <p>
	 * <b>NOTE</b> map may be null. If it is then the current open map will be
	 * used (see {@link #getActiveMap()} or a new map will be created if that is
	 * null.
	 * </p>
	 * 
	 * @param map
	 *            the map to add the layers to. If null the current active map
	 *            will be used or a new one will be created
	 * @param resourceList
	 *            Resources to add to the map.
	 * @param startPosition
	 *            z-position of the layers to add. if -1 it will be added to the
	 *            top of the map (0 is the bottom of the map and
	 *            map.getMapLayer.size() is the top of the map).
	 * @return layers that were added.
	 */
	public static List<? extends ILayer> addLayersToMap(IMap map,
			List<IGeoResource> resourceList, int startPosition) {
		return addLayersToMap(map, resourceList, startPosition, null, false);
	}

	/**
	 * Make layers from the resourceList, creates a new map, adds layers to map
	 * and adds map to the project.
	 * 
	 * @param project
	 *            project that new map should be added to
	 * @param resourceList
	 *            Resources to add to the map.
	 * @param startPosition
	 *            z-position of the layers to add. if -1 it will be added to the
	 *            top of the map (0 is the bottom of the map and
	 *            map.getMapLayer.size() is the top of the map).
	 * 
	 * @return layers that were added.
	 */
	public static List<? extends ILayer> addLayersToMap(IProject project,
			List<IGeoResource> resourceList) {
		return addLayersToMap(null, resourceList, 0, project, false);
	}

	/**
	 * Make layers from the resourceList and adds the layers to the map.
	 * <p>
	 * <b>NOTE</b> map may be null. If it is then the current open map will be
	 * used (see {@link #getActiveMap()} or a new map will be created if that is
	 * null.
	 * </p>
	 * 
	 * @param map
	 *            the map to add the layers to. If null the current active map
	 *            will be used or a new one will be created
	 * @param resourceList
	 *            Resources to add to the map.
	 * @param startPosition
	 *            z-position of the layers to add. if -1 it will be added to the
	 *            top of the map (0 is the bottom of the map and
	 *            map.getMapLayer.size() is the top of the map).
	 * @param project
	 *            project that map should be added to... Only used if there is
	 *            no current map. If project is then the default project is
	 *            used.
	 * @return layers that were added.
	 * @deprecated
	 */
	public static List<? extends ILayer> addLayersToMap(IMap map,
			List<IGeoResource> resourceList, int startPosition, Project project) {
		return addLayersToMap(map, resourceList, startPosition, project, false);
	}

	/**
	 * Make layers from the resourceList and adds the layers to the map.
	 * <p>
	 * <b>NOTE</b> map may be null. If it is then the current open map will be
	 * used (see {@link #getActiveMap()} or a new map will be created if that is
	 * null.
	 * </p>
	 * 
	 * @param map
	 *            the map to add the layers to. If null the current active map
	 *            will be used or a new one will be created
	 * @param resourceList
	 *            Resources to add to the map.
	 * @param startPosition
	 *            z-position of the layers to add. if -1 it will be added to the
	 *            top of the map (0 is the bottom of the map and
	 *            map.getMapLayer.size() is the top of the map).
	 * @param project
	 *            project that map should be added to... Only used if there is
	 *            no current map. If project is then the default project is
	 *            used.
	 * @param wait
	 *            if true then method will block until map has been opened
	 *            otherwise will return without blocking.
	 */
	public static List<? extends ILayer> addLayersToMap(IMap map2,
			List<IGeoResource> resourceList, int startPosition2,
			IProject project2, boolean wait) {
		IMap map = map2;
		if (map == null && project2 == null)
			map = (IMap) getActiveMap();
		IProject project = project2;
		if (project == null) {
			if (map == null)
				project = ProjectPlugin.getPlugin().getProjectRegistry()
						.getCurrentProject();
			else
				project = map.getProject();
		}

		List<? extends ILayer> layers;
		if (map == null) {
			CreateMapCommand cmCommand = new CreateMapCommand(null,
					resourceList, project);
			project.sendSync(cmCommand);
			map = cmCommand.getCreatedMap();
			layers = map.getMapLayers();
		} else {
			AddLayersCommand alCommand = new AddLayersCommand(resourceList,
					startPosition2);
			map.sendCommandSync(alCommand);
			layers = alCommand.getLayers();
		}

		if (!ApplicationGISInternal.getOpenMaps().contains(map)) {
			openMap(map, wait);
		}
		return layers;
	}

	/**
	 * Loads the project element indicated by the url and adds the map to the
	 * provided project.
	 * 
	 * @param url
	 *            the project element to load
	 * @param project
	 *            the project to add the project element to.
	 * @return returns the loaded project element.
	 * @throws IOException
	 *             thrown if there is a problem reading the project element file
	 * @throws IllegalArgumentException
	 *             thrown if the file indicated by the URL is not a project
	 *             element file.
	 */
	public static IProjectElement loadProjectElement(URL url, IProject project)
			throws IOException, IllegalArgumentException {
		URI uri = URI.createURI(url.toString());

		Resource mapResource;
		try {
			mapResource = ProjectRegistryImpl.getProjectRegistry().eResource()
					.getResourceSet().getResource(uri, true);
		} catch (Exception e) {
			throw new IOException(Messages.ApplicationGIS_loadError + uri);
		}

		Object obj;
		try {
			obj = mapResource.getContents().get(0);
		} catch (Exception e) {
			throw new IllegalArgumentException(
					Messages.ApplicationGIS_illegalArgumentPart1 + uri
							+ Messages.ApplicationGIS_illegalArgumentPart2);
		}
		if (!(obj instanceof ProjectElement))
			throw new IllegalArgumentException(
					Messages.ApplicationGIS_noProjectElement
							+ obj.getClass().getSimpleName());

		ProjectElement elem = (ProjectElement) obj;
		((Project) project).getElementsInternal().add(elem);

		return elem;
	}
	

	/**
	 * Opens a {@link IProjectElement} for editing/viewing.
	 * 
	 * @param obj
	 *            object to open
	 * @param wait
	 *            whether or not to perform the action asynchronously
	 */
	public static void openProjectElement(IProjectElement obj, boolean wait) {
		OpenProjectElementCommand command = new OpenProjectElementCommand(obj);
		if (wait)
			ApplicationGIS.getActiveProject().sendSync(command);
		else
			ApplicationGIS.getActiveProject().sendASync(command);
	}

    /**
     * Parameter class for {@link #drawMap(net.refractions.udig.project.ui.ApplicationGIS.DrawMapParameter)}
     *  
     * @author Jesse
     */
    public static class DrawMapParameter {
        final Graphics2D graphics;
        final Dimension destinationSize;
        final IMap toDraw;
        final int scaleDenominator;
        final int dpi;
        final IProgressMonitor monitor;
        final SelectionStyle selectionStyle;
        
        public DrawMapParameter(Graphics2D graphics, Dimension destinationSize,
                IMap toDraw, int scaleDenominator, int dpi, SelectionStyle selectionStyle,
                IProgressMonitor monitor) {
            this.graphics = graphics;
            this.destinationSize = destinationSize;
            this.toDraw = toDraw;
            this.scaleDenominator = scaleDenominator;
            this.dpi = dpi;
            this.monitor = monitor;
            this.selectionStyle=selectionStyle;
        }

        public DrawMapParameter(Graphics2D graphics, Dimension destinationSize,
                IMap toDraw, IProgressMonitor monitor) {
            this(graphics, destinationSize, toDraw, -1, 72, SelectionStyle.OVERLAY, monitor);
        }
    }
    
    /**
     * Describes how to handle a layer's selection.
     * 
     * @author Jesse
     */
    public enum SelectionStyle{
        /**
         * Don't draw the selections
         */
        IGNORE, 
        /**
         * Overlay the selection on top of the normal rendering 
         */
        OVERLAY, 
        /**
         * Only draw the selection
         */
        EXCLUSIVE, 
        /**
         * Draw the selection when there is a selection, draw as normal
         */
        EXCLUSIVE_ALL;
    }

    /**
     * Renders the provided map on to the provided graphics2D object.
     * 
     * @param graphics
     *            graphics to draw the map on
     * @param destinationSize
     *            the size of the area to draw the map to
     * @param toDraw
     *            the map to draw
     * @param scaleDenominator
     *            the scaleDenominator to print the map at.  If -1 the current zoom level will be used
     * @param monitor
     *            a progress monitor.
     * @throws RenderException
     *             Thrown if an error occurs such as a unreachable server.
     *             
     * @return the map that was rendered.  It will not be saved or and is not part of any project.
     * @deprecated Use {@link #drawMap(DrawMapParameter)} instead
     */
    public static IMap drawMap(final Graphics2D graphics,
            final Dimension destinationSize, final IMap toDraw, final int scaleDenominator,
            final int dpi, IProgressMonitor monitor) throws RenderException {
                return drawMap(new DrawMapParameter(graphics, destinationSize,
                        toDraw, scaleDenominator, dpi, SelectionStyle.OVERLAY, monitor));
            }

    /**
	 * Renders the provided map on to the provided graphics2D object.
     * @param params parameters that describe how and where to draw the map
	 * 
	 * @throws RenderException
	 *             Thrown if an error occurs such as a unreachable server.
	 *             
	 * @return the map that was rendered.  It will not be saved or and is not part of any project.
	 */
    public static IMap drawMap(final DrawMapParameter params) throws RenderException {
    	IProgressMonitor monitor = params.monitor;
        final Map map = (Map) EcoreUtil.copy((EObject) params.toDraw);
        for (int i = 0; i < map.getMapLayers().size(); i++) {
            ILayer source = params.toDraw.getMapLayers().get(i);
            Layer dest = map.getLayersInternal().get(i);
            dest.setFilter(source.getFilter());
        }
        
        IRunnableWithProgress runnable = new IRunnableWithProgress() {

            public void run(IProgressMonitor monitor)
                    throws InvocationTargetException, InterruptedException {
                
                List<Layer> overlay = handleSelection(map.getLayersInternal(), params.selectionStyle);
                ProjectUIPlugin
                        .trace(
                                ApplicationGIS.class,
                                "ApplicationGIS.drawMap() beginning rendering of map '" + map.getName() + "'", null); //$NON-NLS-1$ //$NON-NLS-2$

                RenderManager manager = RenderFactory.eINSTANCE
                        .createRenderManagerViewer();

                map.setRenderManagerInternal(manager);
                
                RenderContext tools = new RenderContextImpl();
                tools.setMapInternal(map);
                tools.setRenderManagerInternal(manager);

                ProjectUIPlugin
                        .trace(
                                ApplicationGIS.class,
                                "Firing size changed event. Changing to size: " + params.destinationSize.width + " by " + params.destinationSize.height, null); //$NON-NLS-1$ //$NON-NLS-2$

                manager.setMapInternal(map);
                manager.setMapDisplay(new IMapDisplay() {
                    public java.awt.Dimension getDisplaySize() {
                        return new java.awt.Dimension(params.destinationSize.width,
                                params.destinationSize.height);
                    }

                    public int getWidth() {
                        return params.destinationSize.width;
                    }

                    public int getHeight() {
                        return params.destinationSize.height;
                    }

                    public int getDPI() {
                        return params.dpi;
                    }
                });

                ViewportModel model = RenderFactory.eINSTANCE
                        .createViewportModel();

                manager.setViewportModelInternal(model);

                model.setCRS(map.getViewportModel().getCRS());
                model.zoomToBox(map.getViewportModel().getBounds());

                ProjectUIPlugin
                        .trace(
                                ApplicationGIS.class,
                                "Using bounds: " + map.getViewportModel().getBounds(), null); //$NON-NLS-1$

                model.setMapInternal(map);

                model
                        .sizeChanged(new MapDisplayEvent(null,
                                new java.awt.Dimension(0, 0),
                                new java.awt.Dimension(params.destinationSize.width,
                                        params.destinationSize.height)));

                
                if( params.scaleDenominator!=-1 ){
                	model.setScale(params.scaleDenominator);
                }

                RendererCreator decisive = new RendererCreatorImpl();
                decisive.setContext(tools);

                decisive.getLayers().addAll(overlay);
                decisive.getLayers().addAll(map.getLayersInternal());

                SortedSet<RenderContext> sortedContexts = new TreeSet<RenderContext>(
                        decisive.getConfiguration());

                render(params, monitor, decisive, sortedContexts);

            }

            private List<Layer> handleSelection(List<Layer> layersInternal,
                    SelectionStyle selectionStyle) {
                
                ArrayList<Layer> overlay = new ArrayList<Layer>();
                for (Layer layer : layersInternal) {
                    Filter selectionFilter = layer.getFilter();
                    switch (selectionStyle) {
                    case EXCLUSIVE:
                        if( selectionFilter!=Filter.EXCLUDE){
                            layer.getStyleBlackboard().put(ShowViewInterceptor.KEY, selectionFilter);
                        } else {
                            layer.setVisible(false);
                        }
                        break;
                    case EXCLUSIVE_ALL:
                        if( selectionFilter!=Filter.EXCLUDE){
                            layer.getStyleBlackboard().put(ShowViewInterceptor.KEY, selectionFilter);
                        }
                        break;
                    case OVERLAY:
                        overlay.add(0, new SelectionLayer(layer));
                        break;

                    default:
                        break;
                    }
                }
                return overlay;
            }

            private void render(final DrawMapParameter params,
                    IProgressMonitor monitor, RendererCreator decisive,
                    SortedSet<RenderContext> sortedContexts)
                    throws InvocationTargetException {
                Iterator<RenderContext> iter = sortedContexts.iterator();
                while (iter.hasNext()) {
                    RenderContext context = (RenderContext) iter.next();
                    if (context instanceof CompositeRenderContext) {
                        CompositeRenderContext compositeContext = (CompositeRenderContext) context;
                        List<ILayer> layers = compositeContext.getLayers();
                        boolean visible = false;
                        for (ILayer layer : layers) {
                            visible = visible || layer.isVisible();
                        }
                        if (!visible)
                            continue;
                    } else if (!context.getLayer().isVisible())
                        continue;
                    Renderer renderer = decisive.getRenderer(context);
                    ProjectUIPlugin
                            .trace(
                                    ApplicationGIS.class,
                                    "Issuing render call to " + renderer.getName(), null); //$NON-NLS-1$
                    try {
                        renderer.render(params.graphics, monitor);
                    } catch (RenderException e) {
                        throw new InvocationTargetException(e);
                    }
                }
            }

        };

        try {
            PlatformGIS.runBlockingOperation(runnable, monitor);
        } catch (InvocationTargetException e) {
            throw (RenderException) e.getCause();
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }

        return map;
    }
    
}
